unit UninstProgressDlg;

{
  Inno Setup
  Copyright (C) 1998-2000 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Uninstaller's progress window
}

interface

uses
  Windows;

procedure CreateUninstProgressDlg (const Title, AppName: String; const ModernWizardStyle: Boolean);
procedure UpdateProgressDlg (Progress, Range: Integer);

var
  ProgressDlg: HWND;
  ProgressDlgUpdateTimerExpired: Boolean;
  ProgressDlgCloseTimerExpired: Boolean;

implementation

{$R *.RES}

uses
  Messages, CommCtrl, SysUtils, Msgs, MsgIDs, WinDlgs;

const
  IDC_BODYLABEL                 = 1000;
  IDC_PROGRESS                  = 1001;
  IDC_HEADERRECT                = 1002;
  IDC_HEADERLABEL               = 1003;
  IDC_HEADERDESCRIPTION         = 1004;
  IDC_ICON                      = 1005;
  IDC_BEVELEDLABEL              = 1006;
  IDC_CANCELBUTTON              = 1007;

var
  LastRange: Integer;

type
  PDlgData = ^TDlgData;
  TDlgData = record
    Title, AppName: String;
  end;

procedure AdjustBounds(hwndControl: HWND; Caption: String);
var
  DC: HDC;
  SaveFont: HFont;
  TextSize: TSize;
begin
  { Non-VCL version of TStaticText.AdjustBounds }
  DC := GetDC(0);
  SaveFont := SelectObject(DC, SendMessage(hwndControl, WM_GETFONT, 0, 0));
  GetTextExtentPoint32(DC, PChar(Caption), Length(Caption), TextSize);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  SetWindowPos(hWndControl, 0, 0, 0, TextSize.cx + (GetSystemMetrics(SM_CXBORDER) * 4), TextSize.cy + (GetSystemMetrics(SM_CYBORDER) * 4), SWP_NOMOVE + SWP_NOZORDER + SWP_NOACTIVATE)
end;

function DialogProc (hwndDlg: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): BOOL;
stdcall;
var
  Ps: TPaintStruct;
  Dc: HDC;
  Icon: HICON;
  Rect: TRect;
  Font, OldFont: hFont;
  LogFont: TLogFont;
  Text: array [0..256] of Char;
begin
  Result := False;
  case uMsg of
    WM_INITDIALOG: begin
        if Lo(GetVersion) < 4 then
          { NT 3.x doesn't recognize the DS_CENTER dialog style }
          CenterWnd (hwndDlg);
        SetWindowText (hwndDlg, PChar(PDlgData(lParam).Title));

        if GetDlgItem(hwndDlg, IDC_HEADERLABEL) <> 0 then begin
          OldFont := SendDlgItemMessage(hwndDlg, IDC_HEADERLABEL, WM_GETFONT, 0, 0);
          GetObject(OldFont, SizeOf(LogFont), @LogFont);
          LogFont.lfWeight := FW_BOLD;
          Font := CreateFontIndirect(LogFont);
          SendDlgItemMessage(hwndDlg, IDC_HEADERLABEL, WM_SETFONT, Font, 0);

          SetDlgItemText(hwndDlg, IDC_HEADERLABEL, PChar(SetupMessages[msgWizardUninstalling]));
          SetDlgItemText(hwndDlg, IDC_HEADERDESCRIPTION, PChar(FmtSetupMessage(msgUninstallStatusLabel, [PDlgData(lParam).AppName])));
          SetDlgItemText(hwndDlg, IDC_BODYLABEL, PChar(FmtSetupMessage(msgStatusUninstalling, [PDlgData(lParam).AppName])));
          SetDlgItemText(hwndDlg, IDC_CANCELBUTTON, PChar(SetupMessages[msgButtonCancel]));
          if SetupMessages[msgBeveledLabel] <> '' then begin
            SetDlgItemText(hWndDlg, IDC_BEVELEDLABEL, PChar(' ' + SetupMessages[msgBeveledLabel]));
            AdjustBounds(GetDlgItem(hWndDlg, IDC_BEVELEDLABEL), ' ' + SetupMessages[msgBeveledLabel]);
          end else
            ShowWindow(GetDlgItem(hWndDlg, IDC_BEVELEDLABEL), SW_HIDE);

          { Hide header controls since they are 'placeholders'. See WM_PAINT }
          ShowWindow(GetDlgItem(hwndDlg, IDC_HEADERLABEL), SW_HIDE);
          ShowWindow(GetDlgItem(hwndDlg, IDC_HEADERDESCRIPTION), SW_HIDE);
          ShowWindow(GetDlgItem(hwndDlg, IDC_HEADERRECT), SW_HIDE);
          ShowWindow(GetDlgItem(hwndDlg, IDC_ICON), SW_HIDE);
        end else
          SetDlgItemText (hwndDlg, IDC_BODYLABEL, PChar(FmtSetupMessage(msgUninstallStatusLabel, [PDlgData(lParam).AppName])));

        { Kill the close button }
        DeleteMenu (GetSystemMenu(hwndDlg, False), SC_CLOSE, MF_BYCOMMAND);
        SetTimer (hwndDlg, 98, 125, nil);
        SetTimer (hwndDlg, 99, 1000, nil);
        Result := True;
      end;
    WM_PAINT: begin
        if GetDlgItem(hwndDlg, IDC_HEADERLABEL) <> 0 then begin
          Dc := BeginPaint(hwndDlg, ps);

          GetWindowRect(GetDlgItem(hwndDlg, IDC_HEADERRECT), Rect);
          MapWindowPoints(HWND_DESKTOP, hwndDlg, Rect, 2);
          FillRect(Dc, Rect, COLOR_WINDOW+1);

          GetWindowRect(GetDlgItem(hwndDlg, IDC_HEADERLABEL), Rect);
          MapWindowPoints(HWND_DESKTOP, hwndDlg, Rect, 2);
          GetDlgItemText(hwndDlg, IDC_HEADERLABEL, Text, Sizeof(Text));
          OldFont := SelectObject(DC, SendDlgItemMessage(hwndDlg, IDC_HEADERLABEL, WM_GETFONT, 0, 0));
          SetTextColor(DC, GetSysColor(COLOR_WINDOWTEXT));
          SetBkColor(DC, GetSysColor(COLOR_WINDOW));
          DrawText(Dc, Text, -1, Rect, DT_LEFT or DT_NOPREFIX or DT_WORDBREAK);
          SelectObject(Dc, OldFont);

          GetWindowRect(GetDlgItem(hwndDlg, IDC_HEADERDESCRIPTION), Rect);
          MapWindowPoints(HWND_DESKTOP, hwndDlg, Rect, 2);
          GetDlgItemText(hwndDlg, IDC_HEADERDESCRIPTION, Text, Sizeof(Text));
          OldFont := SelectObject(DC, SendDlgItemMessage(hwndDlg, IDC_HEADERDESCRIPTION, WM_GETFONT, 0, 0));
          SetTextColor(DC, GetSysColor(COLOR_WINDOWTEXT));
          SetBkColor(DC, GetSysColor(COLOR_WINDOW));
          DrawText(Dc, Text, -1, Rect, DT_LEFT or DT_NOPREFIX or DT_WORDBREAK);
          SelectObject(Dc, OldFont);

          GetWindowRect(GetDlgItem(hwndDlg, IDC_ICON), Rect);
          MapWindowPoints(HWND_DESKTOP, hwndDlg, Rect, 2);
          Icon := SendDlgItemMessage(hwndDlg, IDC_ICON, STM_GETICON, 0, 0);
          DrawIcon(Dc, Rect.Left, Rect.Top, Icon);

          EndPaint(hwndDlg, ps);
          Result := True;
        end;
      end;
    WM_QUERYENDSESSION: begin
        { Prevent the user from logging off or shutting down during uninstallation }
        SetWindowLong (hwndDlg, DWL_MSGRESULT, 0);  { return 'False' }
        Result := True;
      end;
    WM_TIMER: begin
        case wParam of
          98: ProgressDlgUpdateTimerExpired := True;
          99: begin
                KillTimer (hwndDlg, 99);
                ProgressDlgCloseTimerExpired := True;
              end;
        end;
      end;
    WM_DESTROY:  begin
        if GetDlgItem(hwndDlg, IDC_HEADERLABEL) <> 0 then
          DeleteObject(SendDlgItemMessage(hwndDlg, IDC_HEADERLABEL, WM_GETFONT, 0, 0));
        Result := True;
      end;
  end;
end;

procedure CreateUninstProgressDlg (const Title, AppName: String; const ModernWizardStyle: Boolean);
const
  TemplateNames: array [Boolean] of PChar = ('UNINSTPROGRESSDLG', 'UNINSTPROGRESSDLGMODERN');
var
  Data: TDlgData;
begin
  LastRange := -1;
  ProgressDlgUpdateTimerExpired := False;
  ProgressDlgCloseTimerExpired := False;

  InitCommonControls;  { the dialog contains a progress bar control }
  Data.Title := Title;
  Data.AppName := AppName;
  ProgressDlg := CreateDialogWithFont(HInstance, TemplateNames[ModernWizardStyle], 0, @DialogProc, LPARAM(@Data));
  ShowWindow (ProgressDlg, SW_SHOW);
end;

procedure UpdateProgressDlg (Progress, Range: Integer);
begin
  if Range <> LastRange then begin
    SendDlgItemMessage (ProgressDlg, IDC_PROGRESS, PBM_SETRANGE, 0,
      MAKELPARAM(0, Range));
    LastRange := Range;
  end;
  SendDlgItemMessage (ProgressDlg, IDC_PROGRESS, PBM_SETPOS, Progress, 0);
end;

end.
